export class M3u8 {
	Parent=null
	Version=0   // EXT-X-VERSION:version
	MediaSequence=0 // Default 0, #EXT-X-MEDIA-SEQUENCE:sequence
	Segments=[]
	MasterPlaylist=[]
	Keys=[]
	EndList=false        // #EXT-X-ENDLIST
	PlaylistType="" // VOD or EVENT
	TargetDuration=0.0      // #EXT-X-TARGETDURATION:duration
	Map={
		URI:"",
		BYTERANGE:""
	}
	MediaList=[]
}

export class Segment {
	URI=""
	KeyIndex=0
	Title=""  // #EXTINF: duration,<title>
	Duration=0.0 // #EXTINF: duration,<title>
	Length=0  // #EXT-X-BYTERANGE: length[@offset]
	Offset=0  // #EXT-X-BYTERANGE: length[@offset]
	BitRate=0
}

export class MasterPlaylist {
	URI=""
	BandWidth=0
	Resolution=""
	Codecs=""
	ProgramID=0
	Audio=""
	Subtitles=""
}

export class Key {
	// 'AES-128' or 'NONE'
	// If the encryption method is NONE, the URI and the IV attributes MUST NOT be present
	Method=""
	URI=""
	IV=""
}

export class Media {
	Type=""
	GROUP_ID=""
	NAME=""
	AUTOSELECT=""
	FORCED=""
	LANGUAGE=""
	URI=""
}

var linePattern = /([a-zA-Z-]+)=("[^"]+"|[^",]+)/g

export function m3u8_parser(url, content, m3u8)
{
	let lines = content.split("\n")
	let url_folder = getURLFolder(url)

	let extInf = false
	let keyIndex = 0
	let seg = null
	let extByte = false
	let extBit = false

	for (let i=0;i<lines.length;i++) {
		let line = lines[i].trim()
		if (i==0) {
			if ("#EXTM3U" != line) {
				throw new Error('invalid m3u8, missing #EXTM3U in line 1')
			}
		}
		if (line=="") continue;
		
		if (line.startsWith("#EXT-X-PLAYLIST-TYPE:")) {
			line = line.replace("#EXT-X-PLAYLIST-TYPE:", '')
			m3u8.PlaylistType = line
			if (m3u8.PlaylistType != "" && m3u8.PlaylistType != "VOD" && m3u8.PlaylistType != "EVENT") {
				throw new Error("invalid playlist type: "+m3u8.PlaylistType+", line: "+(i+1))
			}
		} else if (line.startsWith("#EXT-X-TARGETDURATION:")) {
			line = line.replace("#EXT-X-TARGETDURATION:", '')
			m3u8.TargetDuration = line
		} else if (line.startsWith("#EXT-X-MEDIA-SEQUENCE:")) {
			line = line.replace("#EXT-X-MEDIA-SEQUENCE:", '')
			m3u8.MediaSequence = line
		} else if (line.startsWith("#EXT-X-VERSION:")) {
			line = line.replace("#EXT-X-VERSION:", "")
			m3u8.Version = line
		} else if (line.startsWith("#EXT-X-STREAM-INF:")) {
			let mp = parseMasterPlaylist(line)
			i++
			mp.URI = lines[i]
			if (mp.URI == "" || mp.URI.startsWith("#")) {
				throw new Error("invalid EXT-X-STREAM-INF URI, line:"+(i+1))
			}
			m3u8.MasterPlaylist.push(mp)
		} else if (line.startsWith("#EXT-X-MAP:")) {
			line = line.replace('#EXT-X-MAP:', '')
			let params = parseLineParameters(line)
			if (Object.keys(params).length==0) {
				throw new Error("invalid EXT-X-MAP: "+line+", line: "+(i+1))
			}
			m3u8.Map.URI = params['URI']
			if (m3u8.Map.URI.indexOf('://')==-1) {
				if (m3u8.Map.URI.startsWith('/')) {
					m3u8.Map.URI = m3u8.Map.URI.substring(1)
				}
				m3u8.Map.URI = url_folder+m3u8.Map.URI
			}
		} else if (line.startsWith("#EXT-X-MEDIA:")) {
			line = line.replace("#EXT-X-MEDIA:", '')
			let params = parseLineParameters(line)
			if (Object.keys(params).length==0) {
				throw new Error("invalid EXT-X-MEDIA: "+line+", line: "+(i+1))
			}
			let med = new Media()
			med.URI = params['URI']
			med.GROUP_ID = params['GROUP-ID']
			med.Type = params['TYPE']
			if (med.URI.indexOf('://')==-1) {
				if (med.URI.startsWith('/')) {
					med.URI = med.URI.substring(1)
				}
				med.URI = url_folder+med.URI
			}
			m3u8.MediaList.push(med)
		} else if (line.startsWith("#EXTINF:")) {
			if (extInf) {
				throw new Error("duplicate EXTINF: "+line+", line: "+(i+1))
			}
			line = line.replace("#EXTINF:", '')
			let s = line
			if (seg == null) {
				seg = new Segment()
			}
			if (s.indexOf(",")>-1) {
				let split = s.split(",")
				seg.Title = s[1]
				s = s[0]
			}
			seg.Duration = s
			seg.KeyIndex = keyIndex
			extInf = true
		} else if (line.startsWith("#EXT-X-BYTERANGE:")) {
			if (extByte) {
				throw new Error("duplicate EXT-X-BYTERANGE: line: "+(i+1))
			}
			if (seg == null) {
				seg = new Segment()
			}
			line = line.replace('#EXT-X-BYTERANGE:', '')
			let b = line
			if (b=='') {
				throw new Error("invalid EXT-X-BYTERANGE: line: "+(i+1))
			}
			if (b.indexOf("@")>-1) {
				split = b.split("@")
				seg.Offset = b[1]
				b= split[0]
			}
			seg.Length = b
			extByte = true
		} else if (line.startsWith("#EXT-X-BITRATE:")) {
			if (extBit) {
				throw new Error("duplicate EXT-X-BITRATE: line: "+(i+1))
			}
			if (seg==null) {
				throw new Error("invalid line:"+(i+1))
			}
			line = line.replace('#EXT-X-BITRATE:', '')
			if (line=='') {
				throw new Error("invalid EXT-X-BITRATE: line: "+(i+1))
			}
			seg.BitRate = line
			extBit = true
		} else if (!line.startsWith('#')) {
			if (extInf) {
				if (seg==null) {
					throw new Error("invalid line:"+(i+1))
				}
				seg.URI = line
				if (seg.URI.indexOf('://')==-1) {
					if (seg.URI.startsWith('/')) {
						seg.URI = seg.URI.substring(1)
					}
					seg.URI = url_folder+seg.URI
				}
				extByte =false
				extInf = false
				extBit = false
				m3u8.Segments.push(seg)
				seg = null
			}
		} else if (line.startsWith("#EXT-X-KEY")) {
			let params = parseLineParameters(line)
			if (Object.keys(params).length==0) {
				throw new Error("invalid EXT-X-KEY: "+line+", line: "+(i+1))
			}
			let method = params["METHOD"]
			if (method != "" && method != "AES-128" && method != "NONE") {
				throw new Error("invalid EXT-X-KEY method: "+method+", line: "+(i+1))
			}
			keyIndex++
			let key = new Key()
			key.Method = method
			key.URI = params["URI"]
			key.IV = params["IV"]
			m3u8.Keys[keyIndex] = key
		} else if (line=="#EXT-X-EndList") {
			m3u8.EndList = true
		}
	}
}

function parseMasterPlaylist(line)
{
	let params = parseLineParameters(line)
	if (Object.keys(params).length == 0) {
		throw new Error("empty parameter")
	}
	let mp = new MasterPlaylist()
	for (let k in params) {
		if (k=="BANDWIDTH") {
			mp.BandWidth = params[k]
		} else if (k=="RESOLUTION") {
			mp.Resolution = params[k]
		} else if (k=="PROGRAM-ID") {
			mp.ProgramID = params[k]
		} else if (k=="CODECS") {
			mp.Codecs = params[k]
		} else if (k=="AUDIO") {
			mp.Audio = params[k]
		} else if (k=='SUBTITLES') {
			mp.Subtitles = params[k]
		}
	}
	return mp
}

function parseLineParameters(line) {
	let params = []
	let matches = line.matchAll(linePattern)
	for (const match of matches) {
		params[match[1]] = match[2].replaceAll('"', '')
	}
	return params
}

export function m3u8_parseUrl(url, parent_m3u8=null)
{
	return fetch(new Request(url)).then((response)=>{
		if (response.status==200) {
			let file = new M3u8()
			file.Parent = parent_m3u8
			return response.text().then((data)=>{
				m3u8_parser(url, data, file);
				if (file.MasterPlaylist.length!=0) {
					return m3u8_parseUrl(file.MasterPlaylist[0].URI, file);
				}
				if (file.Segments.length==0) {
					throw new Error('Can not found any TS file description')
				}
				for (let idx in file.Keys) {
					if (file.Keys[idx].Method=='' || file.Keys[idx].Method=='NONE') {
						continue;
					} else if (file.Keys[idx].Method=='AES-128') {
						let keyURL = file.Keys[idx].URI
						keyURL = ResolveURL(url, keyURL)
					} else {
						throw new Error('Unknown or unsupported cryption method: '+file.Keys[idx].Method)
					}
				}
				return file
			})
		} else {
			throw new Error('Fetch m3u8 file failed')
		}
	})
}

function ResolveURL(url, p) {
	if (p.startsWith("https://") || p.startsWith("http://")) {
		return p
	}
	let baseURL = ''
	let u = url.split('/')
	if (p.indexOf("/") == 0) {
		baseURL = u[0] + "://" + u[2]
	} else {
		baseURL = getURLFolder(url)
	}
	return baseURL + p
}

function getURLFolder(url)
{
	let u = url.split('/')
	let tmp1 = []
	let baseURL = ''
	for (let i=0;i<u.length-1;i++) {
		tmp1.push(u[i])
	}
	baseURL = tmp1.join('/')
	if (baseURL.charAt(baseURL.length-1)!='/') {
		baseURL+='/'
	}
	return baseURL
}